/*
 * 
 * No more Bessel LP filter of Vfb. It runs fine with the
 * nanoverter Vfb opamp LP filter in place
 */

#define ADC_OUT  0          // channel for AC output sample

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define NPWM  200         // 200 for this and 800 for PPWM for 50Hz and 20kHz PWM. Both must be integers.
//#define PPWM  665       // 60.062 Hz - choose one or the other.
#define PPWM  800         // 50.002 Hz 799  49.939 Hz for 800

      
uint8_t uf,oen,v1low, lv_stopped,int_f;
uint16_t pcount;
uint32_t vpwr;
int sst;
uint16_t l[NPWM+1];
float pwr;
uint8_t dbounce;
double Input, Output, Setpoint;
double errSum, lastErr,error,dErr;
double kp, ki, kd;
float xv[5],yv[5];
float ac_output, ac_setpoint;

void init_vars(void);

long pcount_acc;
int pcount_delta, phase_error, old_phase_error;
int pcint,new_delta, p_lock;

uint8_t init_done; 
uint8_t tt=0;

float fppwm;

void setup()
  {
  int i;
  float t,u;

  for(t=0.0,i=0; i <= NPWM; i++,t +=   3.14159/ (float) (NPWM+1))
    {   
    u = 65535.0 * sin(t);   // 1/2 wave sine lookup table, scaled to (16 bits - 1)
    l[i] =  (int)u;  
    }
 
  pinMode(9, OUTPUT);     // V1 drive
  pinMode(10, OUTPUT);    // V2 drive
  pinMode(12, OUTPUT);    // inverter ON/OFF LED, pin 15 on nano, Arduino D12
  pinMode(8,INPUT);       // continuous on/off pin, driven by nano2
  pinMode(5,OUTPUT);      // used for IR2110 shutdown drive, LOW = shutdown
  pinMode(7,OUTPUT);      // DSO sync pulse, each 50Hz
  noInterrupts();

  TCCR1A = _BV(COM1A1) | _BV(COM1B0) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 10;
  ICR1 = PPWM;
  TIMSK1 |= (1 << TOIE1);

 
  cbi(PORTD,4);   // no sync
  cbi(PORTD,5);   // disable gate drive output. data direction flag set by above statement.
  oen = 0;        // output enable. = 1 to run inverter, = 0 to stop inverter
  init_vars();
  p_lock = 0;          // phase lock flag = 0 = NOT locked
  pcount_acc = 0;      // fractional integer math. Each PWM, pcount_delta is added to pcount_acc and divided by 1024
  pcount_delta = 1024; // in effect we are getting an addition of 1. But pcount_delta can be modified, a little bigger or smaller..
  old_phase_error = 0; // used in PID, to obtain derivative
  init_done = 1;
  
  interrupts();             // enable all interrupts
  TIMSK0 = 0;               // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;  

  }


//
// 20Khz SPWM, code has only 50uS to run, takes about 10uS
//
ISR(TIMER1_OVF_vect)        
 {   
  long c;
  ICR1 = (int)fppwm;
  
  if (pcount == 0)
      if (v1low == 1)           // if first 1/2 wave..
      {
       TCCR1A =  _BV(COM1B1) | _BV(WGM11); // config output compare to suit
       cbi(PORTD,7);
      }                        
    else
      {
       TCCR1A =  _BV(COM1A1) | _BV(WGM11);   // config O.C. to suit
       sbi(PORTD,7);
      }
  c = (l[pcount] * vpwr) >> 16;   // scale sine wave by vpwr, 32 bit integer calcs. 16 bit shift fastest of all
  if (v1low == 1)                 // alternate between 2 output compare pins. Get a full 50Hz waveform
      OCR1B = c;                    //  was c 1/2 out this pin
  else
      OCR1A = c;                    // and then 1/2 out this pin. 
  pcount ++;
  if(pcount >= NPWM)          // pcount  will = NPWM, at each 1/2 wave, at the start
    {
    pcount = 0;               // reset counter
    uf=1;                     // enable one PID control loop execution, in loop()
    if (v1low == 1)           // if first 1/2 wave..
      {
       v1low = 0;             // toggle to 2nd half wave
       //sbi(PORTD,7);          //DSO pulse 
      }                        
    else
      {
       v1low = 1;             // toggle to 1st half wave
       //cbi(PORTD,7); 
      }
    }
}


void init_vars()      // called prior to setting oen = 1, as well as from setup()
  {
  ICR1 = PPWM;
  fppwm = (float) PPWM;
  lv_stopped = 0;  
  sst = 0;      // slow start counter 0 - 251
  v1low = 1;    // half wave output flag, 0 or 1
  pcount = 0;   // counter for sine wave lookup table
  uf = 0;       // set to 1 to enable AC output voltage feedback code to execute
  pwr = 0.0;    // PWM duty cycle analog, 0.0 to 1.0
  vpwr= 0;      // integer equivalent of pwr, 0 to PPWM
  dbounce = 0;  // debounce counter for input switch 
  kp = 0.1;     // set up PID for AC output voltage control
  ki = 0.0;
  kd = 0.01;
  errSum = 0.0;        // initialise PID
  lastErr = 0.0;

  cbi(PORTD,4);  
  }


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}


void loop()
{
  float ch0,c4,tf;
  c4 = (float)analogRead(4)/1024.0;   // 0.0 to 1.0;
  tf = (float)PPWM *(1.0 + c4 - 0.5)/1.0;
  //Serial.println(tf);
  fppwm = tf;
  
  // sample channel..
  // In the past, I low pass filtered this. not now, no need thanks to active LP filter on Vfb
  ac_output = (float)analogRead(ADC_OUT) / 1024.0; // scale 0 - 5V to 0.0 to 1.0
 
  if(uf == 1  )    // this will execute at 100Hz
      {            // it runs at near zero volt crossing of AC output       
      uf=0; 
      ac_setpoint = 0.55;       // for same Vfb as EG8010. change AC output voltage via Vfb voltage divider trimpot
      if (oen == 1) sst++;      // slow start. If output is enabled, keep starting up. sst = output pwm% or approx. voltage.
      if (sst > 251)sst = 251;  // stop increasing when sst has got to the end of slow start. when = 251, this enables PID. See below..
      if (oen == 0) sst--;      // if stopping, slow stop
      if (sst <= 0)             // once fully stopped, no more changes to sst needed
        {
        sst = 0;
        cbi(PORTD,5);     // pull IR2184 shutdown LOW to disable gate drive output
        }
      else
        sbi(PORTD,5);     // pull it HIGH, to enable output

      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {                          // sst can increase or decrease, giving the slow start and slow stop.
        if (oen == 1)              // slow stop? I think of it as de-gauss
          { 
          if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID.
          else                    // During slow start need to ensure output is tracking setpoint
            pwr -= 1.0/250.0;     // should output get close enough to matter.
          }
        else
          pwr -= 1.0 /250.0;      // this will execute in slow stop
        }
        
      if (sst == 251)       // once sst = 251, run PID. the above bang-bang control will NOT have executed, because sst = 251
        {                   
        Setpoint = ac_setpoint;     
        Input = ac_output;  
        do_pid();
        pwr = pwr + Output;
        }
        
      if (pwr > 0.98) pwr = 0.98;     // clamp. Do math in floating point to prevent any integer overflow
      if (pwr < 0.01) pwr = 0.01;     // it's slower. We could do clamp after conversion to int but it's your mosfets...
      vpwr = (int)(fppwm * pwr);// finally apply required PWM duty power factor to vpwr     
      check_switch_cont(); 
      }

}

void check_switch_cont()
{   // called at 100 Hz
  if ((PINB & 0x01) == 0x01)     
    {
    if (init_done == 0)
      {
      init_vars();
      init_done = 1;
      }  
    oen = 1;      // slow start
    sbi(PORTB,4); // and light LED on D12, pin 15 of nano
    }

  // scr shutdown - without a dummy connection of D6 to ground, inverter wont run.
  //if (((PINB & 0x01) == 0) || ((PIND & 0x40) == 0x40)) 
  // for testing, do not check scr shutdown pin, just nano2 on/off
  if (((PINB & 0x01) == 0) )       
    {
    init_done = 0;
    oen = 0;    // slow stop
    cbi(PORTB,4); // and blank LED on D12, pin 15 of nano
    }
          
}
